from gwcloud_python.utils.file_upload import check_file
import pytest
from tempfile import NamedTemporaryFile, TemporaryDirectory
from pathlib import Path


def test_get_endpoint_from_uploaded():
    with TemporaryDirectory() as tmp_dir:
        test_file = NamedTemporaryFile(dir=tmp_dir)
        try:
            file_name = check_file(test_file.name)
            assert file_name == Path(test_file.name)
        except Exception:
            pytest.fail("check_file is failing when it shouldn't")

        with pytest.raises(Exception):
            check_file(tmp_dir / 'nonexistant')



from gwcloud_python.utils.file_download import (
    _get_endpoint_from_uploaded,
    _download_files,
    _get_file_map_fn,
    _save_file_map_fn
)
from gwcloud_python.settings import GWCLOUD_FILE_DOWNLOAD_ENDPOINT, GWCLOUD_UPLOADED_JOB_FILE_DOWNLOAD_ENDPOINT
import pytest
from tempfile import TemporaryFile, TemporaryDirectory
from pathlib import Path


@pytest.fixture
def test_file_ids():
    return [
        'test_id_1',
        'test_id_2',
        'test_id_3',
        'test_id_4',
    ]


@pytest.fixture
def test_file_paths():
    return [
        'test_path_1',
        'test_path_2',
        'test_path_3',
        'test_path_4',
    ]


@pytest.fixture
def test_file_uploaded():
    return [
        False,
        False,
        True,
        True,
    ]


@pytest.fixture
def setup_file_download(requests_mock):
    def mock_file_download(test_id, test_path, test_uploaded, test_content):
        test_file = TemporaryFile()
        test_file.write(test_content)
        test_file.seek(0)

        requests_mock.get(_get_endpoint_from_uploaded(test_uploaded) + test_id, body=test_file)
    return mock_file_download


def test_get_endpoint_from_uploaded():
    assert _get_endpoint_from_uploaded(True) == GWCLOUD_UPLOADED_JOB_FILE_DOWNLOAD_ENDPOINT
    assert _get_endpoint_from_uploaded(False) == GWCLOUD_FILE_DOWNLOAD_ENDPOINT


def test_download_files(mocker, test_file_ids, test_file_paths, test_file_uploaded):
    mock_map_fn = mocker.Mock()
    mock_progress = mocker.patch('gwcloud_python.utils.file_download.tqdm')

    _download_files(mock_map_fn, test_file_ids, test_file_paths, test_file_uploaded, 100)
    mock_calls = [
            mocker.call(test_id, test_path, test_uploaded, progress_bar=mock_progress())
            for test_id, test_path, test_uploaded in zip(test_file_ids, test_file_paths, test_file_uploaded)
        ]

    mock_map_fn.assert_has_calls(mock_calls)


def test_get_file_map_fn(setup_file_download, mocker):
    test_id = 'test_id'
    test_path = 'test_path'
    test_uploaded = False
    test_content = b'Test file content'
    setup_file_download(test_id, test_path, test_uploaded, test_content)
    _, file_data = _get_file_map_fn(
        file_id=test_id,
        file_path=test_path,
        file_uploaded=test_uploaded,
        progress_bar=mocker.Mock(),
    )

    assert file_data == test_content


def test_save_file_map_fn(setup_file_download, mocker):
    with TemporaryDirectory() as tmp_dir:
        test_id = 'test_id'
        test_path = Path(tmp_dir) / 'test_path'
        test_uploaded = False
        test_content = b'Test file content'
        setup_file_download(test_id, test_path, test_uploaded, test_content)
        _save_file_map_fn(
            file_id=test_id,
            file_path=test_path,
            file_uploaded=test_uploaded,
            progress_bar=mocker.Mock(),
        )

        with open(test_path, 'rb') as f:
            file_data = f.read()
            assert file_data == test_content



GWCLOUD_ENDPOINT = 'https://gwcloud.org.au/bilby/graphql'
GWCLOUD_FILE_DOWNLOAD_ENDPOINT = 'https://gwcloud.org.au/job/apiv1/file/?fileId='
GWCLOUD_UPLOADED_JOB_FILE_DOWNLOAD_ENDPOINT = 'https://gwcloud.org.au/bilby/file_download/?fileId='



from functools import partial
from pathlib import Path
import re


def write_file_at_path(root_path, file_path, file_contents, preserve_directory_structure=True):
    """Write a file at the given path, with the given contents

    Parameters
    ----------
    root_path : str or pathlib.Path
        The base directory
    file_path : pathlib.Path
        The file path within the base directory, including the name of the file
    file_contents : bytes
        The contents of the file
    preserve_directory_structure : bool, optional
        Create any directories present in `file_path`, by default True
    """
    if preserve_directory_structure:
        path = root_path / file_path
    else:
        path = root_path / Path(file_path.name)

    path.parents[0].mkdir(parents=True, exist_ok=True)
    path.write_bytes(file_contents)


def remove_path_anchor(path):
    """Removes the path anchor, making it a relative path

    Parameters
    ----------
    path : pathlib.Path
        Path from which to strip anchor

    Returns
    -------
    Path
        Relative path
    """
    if path.is_absolute():
        return path.relative_to(path.anchor)
    else:
        return path


def to_snake_case(key):
    """Rewrites a camelCase string in snake_case

    Parameters
    ----------
    key : str
        Key to convert

    Returns
    -------
    str
        Key in snake_case
    """
    return re.sub('([A-Z]+)', r'_\1', key).lower()


def to_camel_case(key):
    """Rewrites a snake_case string in camelCase

    Parameters
    ----------
    key : str
        Key to convert

    Returns
    -------
    str
        Key in camelCase
    """
    return re.sub(r'_([a-z])', lambda m: m.group(1).upper(), key)


def _rename_key(key, key_map={}):
    return key_map.get(key, key)


def _apply_key_funcs(key, funcs):
    for func in funcs:
        key = func(key)
    return key


def recursively_map_dict_keys(obj, func):
    """Recursively traverse dicts or lists of dicts to apply a function to each dictionary key

    Parameters
    ----------
    obj : dict or list
        Object to traverse
    func : function
        Function to apply to dictionary keys

    Returns
    -------
    dict
        Dictionary with keys modified by `func`
    """
    if isinstance(obj, dict):  # if dict, apply to each key
        return {func(k): recursively_map_dict_keys(v, func) for k, v in obj.items()}
    elif isinstance(obj, list):  # if list, apply to each element
        return [recursively_map_dict_keys(elem, func) for elem in obj]
    else:
        return obj


def rename_dict_keys(input_dict, key_map):
    """Renames the keys in a dictionary

    Parameters
    ----------
    input_dict : dict
        Dictionary for which to change the keys
    key_map : dict
        Dictionary which specifies old keys to be swapped with new keys in the input_dict, e.g `{'old_key': 'new_key'}`

    Returns
    -------
    dict
        Copy of `input_dict` with old keys subbed for new keys
    """
    funcs = [partial(_rename_key, key_map=key_map)]
    return recursively_map_dict_keys(input_dict, partial(_apply_key_funcs, funcs=funcs))


def convert_dict_keys(input_dict, key_map={}, reverse=False):
    """Convert the keys of a dictionary from camelCase to snake_case

    Parameters
    ----------
    input_dict : dict
        Dictionary for which to convert the keys
    key_map : dict, optional
        Dictionary which specifies old keys to be swapped with new keys in `input_dict`,
        e.g `{'old_key': 'new_key'}`, by default {}
    reverse : bool, optional
        If True, will return snake_case keys to camelCase, by default False

    Returns
    -------
    dict
        Copy of `input_dict` with keys converted from camelCase to snake_case, and optional other key sets exchanged
    """
    funcs = []
    if key_map:
        funcs.append(partial(_rename_key, key_map=key_map))

    funcs.append(to_camel_case if reverse else to_snake_case)

    return recursively_map_dict_keys(input_dict, partial(_apply_key_funcs, funcs=funcs))



import pytest
from gwcloud_python import BilbyJob, FileReference, FileReferenceList, EventID
from gwcloud_python.utils import file_filters


@pytest.fixture
def png_data_result():
    return FileReferenceList([
        FileReference(
            path='data/dir/test1.png',
            file_size='1',
            download_token='test_token_1',
            job_id='id'
        ),
        FileReference(
            path='data/dir/test2.png',
            file_size='1',
            download_token='test_token_2',
            job_id='id'
        ),
        FileReference(
            path='result/dir/test1.png',
            file_size='1',
            download_token='test_token_3',
            job_id='id'
        ),
        FileReference(
            path='result/dir/test2.png',
            file_size='1',
            download_token='test_token_4',
            job_id='id'
        ),
    ])


@pytest.fixture
def png_extra():
    return FileReferenceList([
        FileReference(
            path='test1.png',
            file_size='1',
            download_token='test_token_5',
            job_id='id'
        ),
        FileReference(
            path='test2.png',
            file_size='1',
            download_token='test_token_6',
            job_id='id'
        ),
        FileReference(
            path='arbitrary/dir/test1.png',
            file_size='1',
            download_token='test_token_7',
            job_id='id'
        ),
        FileReference(
            path='arbitrary/dir/test2.png',
            file_size='1',
            download_token='test_token_8',
            job_id='id'
        ),
    ])


@pytest.fixture
def corner():
    return FileReferenceList([
        FileReference(
            path='test1_corner.png',
            file_size='1',
            download_token='test_token_9',
            job_id='id'
        ),
        FileReference(
            path='test2_corner.png',
            file_size='1',
            download_token='test_token_10',
            job_id='id'
        ),
    ])


@pytest.fixture
def config():
    return FileReferenceList([
        FileReference(
            path='a_config_complete.ini',
            file_size='1',
            download_token='test_token_11',
            job_id='id'
        ),
    ])


@pytest.fixture
def json():
    return FileReferenceList([
        FileReference(
            path='result/dir/a_merge_result.json',
            file_size='1',
            download_token='test_token_12',
            job_id='id'
        ),
    ])


@pytest.fixture
def index():
    return FileReferenceList([
        FileReference(
            path='index.html',
            file_size='1',
            download_token='test_token_13',
            job_id='id'
        ),
    ])


@pytest.fixture
def png(png_data_result, png_extra, corner):
    return png_data_result + png_extra + corner


@pytest.fixture
def default(png_data_result, config, json, index):
    return png_data_result + config + json + index


@pytest.fixture
def full(png, config, json, index):
    return png + config + json + index


@pytest.fixture
def mock_bilby_job(mocker):
    def bilby_job(methods):
        config_dict = {f'{key}.return_value': value for key, value in methods.items()}
        return BilbyJob(
            client=mocker.Mock(**config_dict),
            job_id='test_id',
            name='TestName',
            description='Test description',
            user='Test User',
            event_id={'event_id': 'GW123456'},
            job_status={
                'name': 'Completed',
                'date': '2021-12-02'
            },
        )

    return bilby_job


@pytest.fixture
def mock_bilby_job_files(mock_bilby_job, full):
    return mock_bilby_job({'_get_files_by_job_id': (full, {'bilbyResultFiles': {'isUploadedJob': False}})})


@pytest.fixture
def mock_bilby_job_update(mock_bilby_job):
    return mock_bilby_job({'request': {'updateBilbyJob': {'result': 'Success'}}})


@pytest.fixture
def update_query():
    return """
            mutation BilbyJobEventIDMutation($input: UpdateBilbyJobMutationInput!) {
                updateBilbyJob(input: $input) {
                    result
                }
            }
        """


def test_bilby_job_full_file_list(mock_bilby_job_files, full):
    bilby_job = mock_bilby_job_files
    assert bilby_job.get_full_file_list() == full

    bilby_job.client._get_files_by_job_id.assert_called_once()


def test_bilby_job_file_filters(mocker, mock_bilby_job_files, full, default, png, corner, config):
    bilby_job = mock_bilby_job_files
    assert file_filters.sort_file_list(bilby_job.get_default_file_list()) == file_filters.sort_file_list(default)
    assert file_filters.sort_file_list(bilby_job.get_png_file_list()) == file_filters.sort_file_list(png)
    assert file_filters.sort_file_list(bilby_job.get_corner_plot_file_list()) == file_filters.sort_file_list(corner)
    assert file_filters.sort_file_list(bilby_job.get_config_file_list()) == file_filters.sort_file_list(config)

    assert bilby_job.client._get_files_by_job_id.call_count == 4


def test_register_file_list_filter(mock_bilby_job_files, index):
    bilby_job = mock_bilby_job_files

    def get_html_file(file_list):
        return [f for f in file_list if f.path.suffix == '.html']

    assert getattr(bilby_job, 'get_index_file_list', None) is None
    assert getattr(bilby_job, 'get_index_files', None) is None
    assert getattr(bilby_job, 'save_index_files', None) is None

    BilbyJob.register_file_list_filter('index', get_html_file)

    assert getattr(bilby_job, 'get_index_file_list', None) is not None
    assert getattr(bilby_job, 'get_index_files', None) is not None
    assert getattr(bilby_job, 'save_index_files', None) is not None

    assert bilby_job.get_index_file_list() == index


def test_bilbyjob_set_name(mock_bilby_job_update, update_query):
    bilby_job = mock_bilby_job_update

    assert bilby_job.name == 'TestName'
    bilby_job.set_name(name='ADifferentName')
    assert bilby_job.name == 'ADifferentName'
    bilby_job.client.request.assert_called_once_with(
        query=update_query,
        variables={
            'input': {
                'jobId': bilby_job.job_id,
                'name': 'ADifferentName'
            }
        }
    )


def test_bilbyjob_set_description(mock_bilby_job_update, update_query):
    bilby_job = mock_bilby_job_update

    assert bilby_job.description == 'Test description'
    bilby_job.set_description(description='A different description')
    assert bilby_job.description == 'A different description'
    bilby_job.client.request.assert_called_once_with(
        query=update_query,
        variables={
            'input': {
                'jobId': bilby_job.job_id,
                'description': 'A different description'
            }
        }
    )


def test_bilbyjob_set_event_id(mock_bilby_job, update_query):
    bilby_job = mock_bilby_job({
        'request': {'updateBilbyJob': {'result': 'Success'}},
        'get_event_id': EventID(event_id='GW111111')
    })

    assert bilby_job.event_id == EventID(event_id='GW123456')
    bilby_job.set_event_id(event_id='GW111111')
    assert bilby_job.event_id == EventID(event_id='GW111111')
    bilby_job.client.request.assert_called_once_with(
        query=update_query,
        variables={
            'input': {
                'jobId': bilby_job.job_id,
                'eventId': 'GW111111'
            }
        }
    )



import pytest
from gwcloud_python.utils import to_snake_case, to_camel_case, convert_dict_keys, rename_dict_keys


@pytest.fixture
def snake_case():
    return [
        'single',
        'short_string',
        'a_much_longer_test_string',
        'test_key'
    ]


@pytest.fixture
def camel_case():
    return [
        'single',
        'shortString',
        'aMuchLongerTestString',
        'testKey'
    ]


@pytest.fixture
def snake_case_dict(snake_case):
    return {
        snake_case[0]: 0,
        snake_case[1]: 1,
        snake_case[2]: 2,
        snake_case[3]: [
            snake_case[0],
            {snake_case[2]: 2}
        ],
        snake_case[3]: {
            snake_case[0]: 0,
            snake_case[1]: 1,
            snake_case[2]: 2,
        }
    }


@pytest.fixture
def snake_case_renamed_dict(snake_case):
    renamed_map = {snake_case[2]: 'renamed'}
    renamed_dict = {
        snake_case[0]: 0,
        snake_case[1]: 1,
        'renamed': 2,
        snake_case[3]: [
            snake_case[0],
            {'renamed': 2}
        ],
        snake_case[3]: {
            snake_case[0]: 0,
            snake_case[1]: 1,
            'renamed': 2,
        }
    }
    return renamed_map, renamed_dict


@pytest.fixture
def camel_case_dict(camel_case):
    return {
        camel_case[0]: 0,
        camel_case[1]: 1,
        camel_case[2]: 2,
        camel_case[3]: [
            camel_case[0],
            {camel_case[2]: 2}
        ],
        camel_case[3]: {
            camel_case[0]: 0,
            camel_case[1]: 1,
            camel_case[2]: 2,
        }
    }


@pytest.fixture
def camel_case_renamed_dict(camel_case):
    renamed_map = {camel_case[2]: 'renamed'}
    renamed_dict = {
        camel_case[0]: 0,
        camel_case[1]: 1,
        'renamed': 2,
        camel_case[3]: [
            camel_case[0],
            {'renamed': 2}
        ],
        camel_case[3]: {
            camel_case[0]: 0,
            camel_case[1]: 1,
            'renamed': 2,
        }
    }
    return renamed_map, renamed_dict


def test_to_snake_case(snake_case, camel_case):
    for snake, camel in zip(snake_case, camel_case):
        assert to_snake_case(camel) == snake


def test_to_camel_case(snake_case, camel_case):
    for snake, camel in zip(snake_case, camel_case):
        assert to_camel_case(snake) == camel


def test_renamed_dict_keys(snake_case_dict, snake_case_renamed_dict, camel_case_dict, camel_case_renamed_dict):
    assert rename_dict_keys(snake_case_dict, snake_case_renamed_dict[0]) == snake_case_renamed_dict[1]
    assert rename_dict_keys(camel_case_dict, camel_case_renamed_dict[0]) == camel_case_renamed_dict[1]


def test_convert_dict_keys(snake_case_dict, snake_case_renamed_dict, camel_case_dict, camel_case_renamed_dict):
    converted_snake_case_dict = convert_dict_keys(snake_case_dict, reverse=True)
    converted_camel_case_dict = convert_dict_keys(camel_case_dict)

    renamed_snake_case_dict = convert_dict_keys(snake_case_dict, key_map=snake_case_renamed_dict[0], reverse=True)
    renamed_camel_case_dict = convert_dict_keys(camel_case_dict, key_map=camel_case_renamed_dict[0])

    # Can convert
    assert converted_camel_case_dict == snake_case_dict
    assert converted_snake_case_dict == camel_case_dict

    # Conversion is reversible
    assert convert_dict_keys(converted_camel_case_dict, reverse=True) == camel_case_dict
    assert convert_dict_keys(converted_snake_case_dict) == snake_case_dict

    # Can convert and rename from key map
    assert renamed_camel_case_dict == snake_case_renamed_dict[1]
    assert renamed_snake_case_dict == camel_case_renamed_dict[1]



import pytest
from pathlib import Path
from collections import OrderedDict
from gwcloud_python import FileReference, FileReferenceList
from gwcloud_python.utils import remove_path_anchor


@pytest.fixture
def setup_dicts():
    return [
        {
            'path': 'data/dir/test1.png',
            'file_size': '1',
            'download_token': 'test_token_1',
            'job_id': 'id1',
            'is_uploaded_job': False
        },
        {
            'path': 'data/dir/test2.png',
            'file_size': '1',
            'download_token': 'test_token_2',
            'job_id': 'id1',
            'is_uploaded_job': False
        },
        {
            'path': 'result/dir/test1.txt',
            'file_size': '1',
            'download_token': 'test_token_3',
            'job_id': 'id2',
            'is_uploaded_job': True
        },
        {
            'path': 'result/dir/test2.txt',
            'file_size': '1',
            'download_token': 'test_token_4',
            'job_id': 'id2',
            'is_uploaded_job': True
        },
        {
            'path': 'test1.json',
            'file_size': '1',
            'download_token': 'test_token_5',
            'job_id': 'id3',
            'is_uploaded_job': False
        },
        {
            'path': 'test2.json',
            'file_size': '1',
            'download_token': 'test_token_6',
            'job_id': 'id3',
            'is_uploaded_job': False
        },
    ]


def test_file_reference(setup_dicts):
    for file_dict in setup_dicts:
        ref = FileReference(**file_dict)
        assert ref.path == remove_path_anchor(Path(file_dict['path']))
        assert ref.file_size == int(file_dict['file_size'])
        assert ref.download_token == file_dict['download_token']
        assert ref.job_id == file_dict['job_id']
        assert ref.is_uploaded_job == file_dict['is_uploaded_job']


def test_file_reference_list(setup_dicts):
    file_references = [FileReference(**file_dict) for file_dict in setup_dicts]
    file_reference_list = FileReferenceList(file_references)

    for i, ref in enumerate(file_reference_list):
        assert ref.path == file_references[i].path
        assert ref.file_size == file_references[i].file_size
        assert ref.download_token == file_references[i].download_token
        assert ref.job_id == file_references[i].job_id
        assert ref.is_uploaded_job == file_references[i].is_uploaded_job

    assert file_reference_list.get_total_bytes() == sum([ref.file_size for ref in file_references])
    assert file_reference_list.get_tokens() == [ref.download_token for ref in file_references]
    assert file_reference_list.get_paths() == [ref.path for ref in file_references]
    assert file_reference_list.get_uploaded() == [ref.is_uploaded_job for ref in file_references]


def test_file_reference_list_output_paths(setup_dicts):
    file_reference_list = FileReferenceList([FileReference(**file_dict) for file_dict in setup_dicts])

    root_path = Path('test_dir')
    output_paths = [
        root_path / 'data/dir/test1.png',
        root_path / 'data/dir/test2.png',
        root_path / 'result/dir/test1.txt',
        root_path / 'result/dir/test2.txt',
        root_path / 'test1.json',
        root_path / 'test2.json'
    ]
    output_paths_flat = [
        root_path / 'test1.png',
        root_path / 'test2.png',
        root_path / 'test1.txt',
        root_path / 'test2.txt',
        root_path / 'test1.json',
        root_path / 'test2.json'
    ]
    assert output_paths == file_reference_list.get_output_paths(root_path)
    assert output_paths_flat == file_reference_list.get_output_paths(root_path, preserve_directory_structure=False)


def test_batch_file_reference_list(setup_dicts):
    file_reference_list = FileReferenceList([FileReference(**file_dict) for file_dict in setup_dicts])

    batched = OrderedDict(
        id1=file_reference_list[0:2],
        id2=file_reference_list[2:4],
        id3=file_reference_list[4:6],
    )

    assert file_reference_list._batch_by_job_id() == batched



from .utils import (
    write_file_at_path,
    remove_path_anchor,
    to_snake_case,
    to_camel_case,
    rename_dict_keys,
    convert_dict_keys,
)
from . import identifiers
from . import file_filters
from . import file_download



from gwcloud_python import GWCloud, FileReference, FileReferenceList, JobStatus, EventID
from gwcloud_python.utils import convert_dict_keys
import pytest
from tempfile import TemporaryFile


@pytest.fixture
def mock_gwdc_init(mocker):
    def mock_init(self, token, endpoint):
        pass

    mocker.patch('gwdc_python.gwdc.GWDC.__init__', mock_init)


@pytest.fixture
def setup_mock_gwdc(mocker, mock_gwdc_init):
    def mock_gwdc(request_data):
        mock_request = mocker.Mock(return_value=request_data)
        mocker.patch('gwdc_python.gwdc.GWDC.request', mock_request)

    return mock_gwdc


@pytest.fixture
def single_job_request(setup_mock_gwdc):
    job_data = {
        "id": 1,
        "name": "test_name",
        "description": "test description",
        "user": "Test User1",
        "eventId": {
            "eventId": "GW123456"
        },
        "jobStatus": {
            "name": "Completed",
            "date": "2021-12-02"
        }
    }
    setup_mock_gwdc({"bilbyJob": job_data})
    return job_data


@pytest.fixture
def multi_job_request(setup_mock_gwdc):
    def modify_query_name(query_name):
        job_data_1 = {
            "id": 1,
            "name": "test_name_1",
            "description": "test description 1",
            "user": "Test User1",
            "eventId": {
                "eventId": "GW123456"
            },
            "jobStatus": {
                "name": "Completed",
                "date": "2021-01-01"
            }
        }

        job_data_2 = {
            "id": 2,
            "name": "test_name_2",
            "description": "test description 2",
            "user": "Test User2",
            "eventId": {
                "eventId": "GW123456"
            },
            "jobStatus": {
                "name": "Completed",
                "date": "2021-02-02"
            }
        }

        job_data_3 = {
            "id": 3,
            "name": "test_name_3",
            "description": "test description 3",
            "user": "Test User3",
            "eventId": {
                "eventId": "GW123456"
            },
            "jobStatus": {
                "name": "Error",
                "date": "2021-03-03"
            }
        }

        setup_mock_gwdc({
            query_name: {
                "edges": [
                    {"node": job_data_1},
                    {"node": job_data_2},
                    {"node": job_data_3},
                ]
            }
        })

        return [job_data_1, job_data_2, job_data_3]

    return modify_query_name


@pytest.fixture
def job_file_request(setup_mock_gwdc):
    job_file_data_1 = {
        "path": "path/to/test.png",
        "fileSize": "1",
        "downloadToken": "test_token_1",
        "isDir": False
    }

    job_file_data_2 = {
        "path": "path/to/test.json",
        "fileSize": "10",
        "downloadToken": "test_token_2",
        "isDir": False
    }

    job_file_data_3 = {
        "path": "path/to/test",
        "fileSize": "100",
        "downloadToken": "test_token_3",
        "isDir": True
    }

    setup_mock_gwdc({
        "bilbyResultFiles": {
            "files": [
                job_file_data_1,
                job_file_data_2,
                job_file_data_3
            ],
            "isUploadedJob": False
        }
    })

    return [job_file_data_1, job_file_data_2, job_file_data_3]


@pytest.fixture
def test_files():
    return FileReferenceList([
        FileReference(
            path='test/path_1.png',
            file_size=1,
            download_token='test_token_1',
            job_id='id1',
            is_uploaded_job=False
        ),
        FileReference(
            path='test/path_2.png',
            file_size=1,
            download_token='test_token_2',
            job_id='id1',
            is_uploaded_job=False
        ),
        FileReference(
            path='test/path_3.png',
            file_size=1,
            download_token='test_token_3',
            job_id='id1',
            is_uploaded_job=False
        ),
        FileReference(
            path='test/path_4.png',
            file_size=1,
            download_token='test_token_4',
            job_id='id2',
            is_uploaded_job=True
        ),
        FileReference(
            path='test/path_5.png',
            file_size=1,
            download_token='test_token_5',
            job_id='id2',
            is_uploaded_job=True
        ),
        FileReference(
            path='test/path_6.png',
            file_size=1,
            download_token='test_token_6',
            job_id='id2',
            is_uploaded_job=True
        ),
    ])


@pytest.fixture
def setup_mock_download_fns(mocker, mock_gwdc_init, test_files):
    mock_files = mocker.Mock(return_value=[(f.path, TemporaryFile()) for f in test_files])

    def get_mock_ids(job_id, tokens):
        return [f'{job_id}{i}' for i, _ in enumerate(tokens)]

    mock_ids = mocker.Mock(side_effect=get_mock_ids)
    return (
        mocker.patch('gwcloud_python.gwcloud._download_files', mock_files),
        mocker.patch('gwcloud_python.gwcloud._get_file_map_fn'),
        mocker.patch('gwcloud_python.gwcloud._save_file_map_fn'),
        mocker.patch('gwcloud_python.gwcloud.GWCloud._get_download_ids_from_tokens', mock_ids)
    )


@pytest.fixture
def user_jobs(multi_job_request):
    return multi_job_request('bilbyJobs')


def test_get_job_by_id(single_job_request):
    gwc = GWCloud(token='my_token')

    job = gwc.get_job_by_id('job_id')

    assert job.job_id == single_job_request["id"]
    assert job.name == single_job_request["name"]
    assert job.description == single_job_request["description"]
    assert job.status == JobStatus(
        status=single_job_request["jobStatus"]["name"],
        date=single_job_request["jobStatus"]["date"]
    )
    assert job.event_id == EventID(**convert_dict_keys(single_job_request["eventId"]))
    assert job.user == single_job_request["user"]


def test_get_user_jobs(user_jobs):
    gwc = GWCloud(token='my_token')

    jobs = gwc.get_user_jobs()

    assert jobs[0].job_id == user_jobs[0]["id"]
    assert jobs[0].name == user_jobs[0]["name"]
    assert jobs[0].description == user_jobs[0]["description"]
    assert jobs[0].status == JobStatus(
        status=user_jobs[0]["jobStatus"]["name"],
        date=user_jobs[0]["jobStatus"]["date"]
    )
    assert jobs[0].event_id == EventID(**convert_dict_keys(user_jobs[0]["eventId"]))
    assert jobs[0].user == user_jobs[0]["user"]

    assert jobs[1].job_id == user_jobs[1]["id"]
    assert jobs[1].name == user_jobs[1]["name"]
    assert jobs[1].description == user_jobs[1]["description"]
    assert jobs[1].status == JobStatus(
        status=user_jobs[1]["jobStatus"]["name"],
        date=user_jobs[1]["jobStatus"]["date"]
    )
    assert jobs[1].event_id == EventID(**convert_dict_keys(user_jobs[1]["eventId"]))
    assert jobs[1].user == user_jobs[1]["user"]

    assert jobs[2].job_id == user_jobs[2]["id"]
    assert jobs[2].name == user_jobs[2]["name"]
    assert jobs[2].description == user_jobs[2]["description"]
    assert jobs[2].status == JobStatus(
        status=user_jobs[2]["jobStatus"]["name"],
        date=user_jobs[2]["jobStatus"]["date"]
    )
    assert jobs[2].event_id == EventID(**convert_dict_keys(user_jobs[2]["eventId"]))
    assert jobs[2].user == user_jobs[2]["user"]


def test_gwcloud_files_by_job_id(job_file_request):
    gwc = GWCloud(token='my_token')

    file_list, is_uploaded_job = gwc._get_files_by_job_id('arbitrary_job_id')

    for i, ref in enumerate(file_list):
        job_file_request[i].pop('isDir')
        assert ref == FileReference(
            **convert_dict_keys(job_file_request[i]),
            job_id='arbitrary_job_id',
            is_uploaded_job=is_uploaded_job,
        )


def test_gwcloud_get_files_by_reference(setup_mock_download_fns, mocker, test_files):
    gwc = GWCloud(token='my_token')
    mock_download_files = setup_mock_download_fns[0]
    mock_get_fn = setup_mock_download_fns[1]
    mock_get_ids = setup_mock_download_fns[3]
    mock_ids = ['id10', 'id11', 'id12', 'id20', 'id21', 'id22']

    files = gwc.get_files_by_reference(test_files)

    mock_calls = [
        mocker.call(job_id, job_files.get_tokens())
        for job_id, job_files in test_files._batch_by_job_id().items()
    ]

    mock_get_ids.assert_has_calls(mock_calls)

    assert [f[0] for f in files] == test_files.get_paths()
    mock_download_files.assert_called_once_with(
        mock_get_fn,
        mock_ids,
        test_files.get_paths(),
        test_files.get_uploaded(),
        test_files.get_total_bytes()
    )


def test_gwcloud_save_batched_files(setup_mock_download_fns, mocker, test_files):
    gwc = GWCloud(token='my_token')
    mock_download_files = setup_mock_download_fns[0]
    mock_save_fn = setup_mock_download_fns[2]
    mock_get_ids = setup_mock_download_fns[3]
    mock_ids = ['id10', 'id11', 'id12', 'id20', 'id21', 'id22']

    gwc.save_files_by_reference(test_files, 'test_dir', preserve_directory_structure=True)

    mock_calls = [
        mocker.call(job_id, job_files.get_tokens())
        for job_id, job_files in test_files._batch_by_job_id().items()
    ]

    mock_get_ids.assert_has_calls(mock_calls)

    mock_download_files.assert_called_once_with(
        mock_save_fn,
        mock_ids,
        test_files.get_output_paths('test_dir', preserve_directory_structure=True),
        test_files.get_uploaded(),
        test_files.get_total_bytes()
    )



import concurrent.futures
from functools import partial
import requests
from tqdm import tqdm
from ..settings import GWCLOUD_FILE_DOWNLOAD_ENDPOINT, GWCLOUD_UPLOADED_JOB_FILE_DOWNLOAD_ENDPOINT


def _get_endpoint_from_uploaded(is_uploaded_job):
    return GWCLOUD_FILE_DOWNLOAD_ENDPOINT \
        if not is_uploaded_job else \
        GWCLOUD_UPLOADED_JOB_FILE_DOWNLOAD_ENDPOINT


def _get_file_map_fn(file_id, file_path, file_uploaded, progress_bar):
    download_url = _get_endpoint_from_uploaded(file_uploaded) + str(file_id)
    content = b''

    with requests.get(download_url, stream=True) as request:
        for chunk in request.iter_content(chunk_size=1024 * 16, decode_unicode=True):
            progress_bar.update(len(chunk))
            content += chunk
    return (file_path, content)


def _save_file_map_fn(file_id, file_path, file_uploaded, progress_bar):
    download_url = _get_endpoint_from_uploaded(file_uploaded) + str(file_id)
    file_path.parents[0].mkdir(parents=True, exist_ok=True)

    with requests.get(download_url, stream=True) as request:
        with file_path.open("wb+") as f:
            for chunk in request.iter_content(chunk_size=1024 * 16):
                progress_bar.update(len(chunk))
                f.write(chunk)


def _download_files(map_fn, file_ids, file_paths, file_uploaded, total_size):
    with concurrent.futures.ThreadPoolExecutor(max_workers=20) as executor:
        progress = tqdm(total=total_size, leave=True, unit='B', unit_scale=True)
        files = list(
            executor.map(
                partial(
                    map_fn,
                    progress_bar=progress
                ),
                file_ids, file_paths, file_uploaded
            )
        )
        progress.close()
    return files
